import datetime

import slurminade
from slurminade import JobBundling

slurminade.update_default_configuration(partition="alg", constraint="alggen02")


@slurminade.slurmify()
def f(hello_world):
    with open("slurminade_example.txt", "a") as f:
        print("hello")
        f.write(hello_world + "\n")


if __name__ == "__main__":
    jid = f.distribute(f"Hello World from slurminade! {datetime.datetime.now()!s}")
    with JobBundling(20) as batch:
        f.distribute("hello 1!")
        f.distribute("hello 2!")



from pathlib import Path

import slurminade

test_file_path = Path("./f_test_file.txt")
print("NOISE123")


@slurminade.slurmify()
def f(s):
    with test_file_path.open("w") as file:
        file.write(s)


def test_create_command_with_noise():
    slurminade.set_entry_point(__file__)
    if test_file_path.exists():
        test_file_path.unlink()
    dispatcher = slurminade.SubprocessDispatcher()
    dispatcher.max_arg_length = 1
    slurminade.set_dispatcher(dispatcher)
    s = "test"
    f.distribute(s)
    assert test_file_path.is_file()
    with test_file_path.open() as file:
        assert file.readline() == s
    test_file_path.unlink(missing_ok=True)



"""
Some security measures to warn you about common mistakes and prevent you from
accidentally DDoSing your slurm environment.

1. Preventing recursive distributions, i.e., slurm nodes also distributing tasks.
2. Limiting the number of distributed tasks.
3. Warn about multiple flushes of batches, often caused by wrong indentation.

You can disable these security mechanisms by
``allow_recursive_distribution``, ``set_dispatch_limit(None)``, and
``disable_warning_for_multiple_flushes``.
"""

import logging
import typing

_exec_flag = False


def on_slurm_node():
    global _exec_flag  # noqa: PLW0602
    return _exec_flag


def guard_recursive_distribution():
    if on_slurm_node():
        msg = """
        You tried to distribute a task recursively. This is not allowed by default,
        because it probably indicates a bug in your code. To save you from accidentally
        overloading your slurm environment, this feature has been disabled by default.
        The most common reason for this error is that you forgot to guard your script
        with 'if __name__==\"__main__\":'. If you are sure that you want to distribute
        tasks recursively, you can disable this security mechanism by calling
        `allow_recursive_distribution` before the first call of `distribute`.
        """
        raise RuntimeError(msg)


def prevent_distribution():
    global _exec_flag  # noqa: PLW0603
    _exec_flag = True


def allow_recursive_distribution() -> None:
    """
    Allow recursive distribution. Dangerous!
    :return: None
    """
    global _exec_flag  # noqa: PLW0603
    _exec_flag = False


class TooManyDispatchesError(RuntimeError):
    def __init__(self, n_calls):
        self.n_calls = n_calls

    def __str__(self):
        return (
            f"Exceeded the dispatch limit of {self.n_calls} calls. "
            f"This limit has been introduced to prevent you from overloading your "
            f"slurm environment in case of a bug. You can increase it"
            f" using `set_dispatch_limit`."
        )


class _DispatchGuard:
    def __init__(self, max_calls):
        self.max_calls = max_calls
        self.remaining_calls = max_calls

    def __call__(self):
        if not self.max_calls:
            return None
        if self.remaining_calls <= 0:
            raise TooManyDispatchesError(self.max_calls)
        self.remaining_calls -= 1
        return self.remaining_calls

    def set_limit(self, n):
        self.max_calls = n
        self.remaining_calls = n


dispatch_guard = _DispatchGuard(100)


def set_dispatch_limit(n: typing.Optional[int]):
    """
    Set a limit to the number of dispatches. This feature has been introduced to
    prevent you from accidentally DDoSing you Slurm environment due to a bug.
    :param n: The maximal number of dispatches.
    :return: None
    """
    dispatch_guard.set_limit(n)


class BatchGuard:
    """
    Warns you if you flush more than once, as putting the flush call in a loop is
    a common mistake, compared to the intended use of flushing once at the end of
    your context, to get the job ids for dependency management.
    """

    already_warned = False

    def __init__(self) -> None:
        self._num_of_flushes = 0

    def _get_error_msg(self) -> str:
        return """
        You repeatedly flushed a batch. There are various scenarios where this
        is done on purpose, but we want to warn you, because it is a common mistake
        to put the flush call in a loop by wrong indentation, instead of calling it
        once at the end of your context. If used for dependency management,
        e.g., with `wait_for`, such a mistake can lead to a faulty execution order.
        This warning allows you to quickly call `scancel -u <username>` to cancel
        your jobs, before they do any harm. If you are sure that you want to flush
        your batch multiple times, you can disable this warning by calling
        `disable_warning_on_repeated_flushes` before the first call of `flush`.
        You can also just ignore this warning, as no action is taken by default.
        """

    def report_flush(self, num_tasks: int) -> None:
        if num_tasks == 0:  # ignore empty flushes
            return
        self._num_of_flushes += 1
        if self._num_of_flushes == 2 and not self.already_warned:
            logging.getLogger("slurminade").warning(self._get_error_msg())
            self.already_warned = True


def disable_warning_on_repeated_flushes():
    """
    Disable the warning on multiple flushes. This is useful if you want to flush
    multiple times in a loop, without getting a warning.
    """
    BatchGuard.already_warned = True



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath("../src"))
autodoc_mock_imports = ["simple_slurm"]

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "slurminade"
copyright = "2023, Dominik Krupke"
author = "Dominik Krupke (TU Braunschweig, IBR, Algorithms Group)"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import datetime

# example with importing a function to be executed by node
from example_2b import f

import slurminade

slurminade.set_dispatch_limit(2)
slurminade.update_default_configuration(partition="alg", constraint="alggen02")


@slurminade.slurmify()
def g(hello_world):
    print("This function 'test' is in", __file__)
    with open("slurminade_example.txt", "w") as f_:
        f(hello_world)
        f_.write(hello_world)


if __name__ == "__main__":
    g.distribute("test f")
    f.distribute(f"Hello World from slurminade! {datetime.datetime.now()!s}")



import socket
import tempfile
import time
from pathlib import Path

import click

from slurminade import join, slurmify, srun


@slurmify()
def _write_to_file(path, content):
    time.sleep(1)
    # get hostname and write it to the file
    hostname = socket.gethostname()
    with Path(path).open("w") as file:
        print("Hello from ", hostname)
        file.write(content + "\n" + hostname)
    # wait a second for the file to be written
    time.sleep(1)


@click.command()
@click.option("--partition", default=None, help="The partition to use.")
@click.option("--constraint", default=None, help="The constraint to use.")
def check_slurm(partition, constraint):
    """
    Check if the code is running on a slurm node.
    """
    # enforce slurm
    from slurminade.conf import update_default_configuration
    from slurminade.dispatcher import SlurmDispatcher, set_dispatcher
    from slurminade.function_map import set_entry_point

    set_dispatcher(SlurmDispatcher())
    print("Setting entry point to ", Path(__file__).resolve())
    set_entry_point(Path(__file__).resolve())

    if partition:
        update_default_configuration(partition=partition)
    if constraint:
        update_default_configuration(constraint=constraint)

    # create a temporary folder for the slurm check
    with tempfile.TemporaryDirectory(dir=".") as tmpdir:
        tmpdir_ = Path(tmpdir).resolve()
        assert tmpdir_.exists()
        # Check 1
        tmp_file_path = tmpdir_ / "check_1.txt"
        _write_to_file.distribute_and_wait(str(tmp_file_path), "test")
        if not Path(tmp_file_path).exists():
            msg = "Slurminade failed: The file was not written to the temporary directory."
            raise Exception(msg)
        with Path(tmp_file_path).open() as file:
            content = file.readlines()
            print(
                "Slurminade check 1 successful. Test was run on node",
                content[1].strip(),
            )

        # Check 2
        tmp_file_path = tmpdir_ / "check_2.txt"
        _write_to_file.distribute(str(tmp_file_path), "test")
        # wait up to 1 minutes for the file to be written
        for _ in range(60):
            if Path(tmp_file_path).exists():
                break
            time.sleep(1)
        if not Path(tmp_file_path).exists():
            msg = "Slurminade failed: The file was not written to the temporary directory."
            raise Exception(msg)
        with Path(tmp_file_path).open() as file:
            content = file.readlines()
            print(
                "Slurminade check 2 successful. Test was run on node",
                content[1].strip(),
            )

        join()

        # Check 3
        tmp_file_path = tmpdir_ / "check_3.txt"
        srun(["touch", str(tmp_file_path)])
        time.sleep(1)
        if not Path(tmp_file_path).exists():
            msg = "Slurminade failed: The file was not written to the temporary directory."
            raise Exception(msg)
        print("Slurminade check 3 successful.")
        tmp_file_path.unlink()

        # Check 4
        tmp_file_path = tmpdir_ / "check_4.txt"
        _write_to_file.distribute_and_wait(str(tmp_file_path), "test")
        if not Path(tmp_file_path).exists():
            msg = "Slurminade failed: The file was not written to the temporary directory."
            raise Exception(msg)
        with Path(tmp_file_path).open() as file:
            content = file.readlines()
            print(
                "Slurminade check 1 successful. Test was run on node",
                content[1].strip(),
            )


if __name__ == "__main__":
    check_slurm()



import typing


class SlurmOptions(dict):
    """
    Primarily just a wrapper to allow using the options in a dict as key.
    Necessary for batching function calls, because only function calls with the
    same options can be bundled.
    """

    def _items(self):
        for k, v in self.items():
            if isinstance(v, dict):
                yield k, SlurmOptions(**v)
            else:
                yield k, v

    def __hash__(self):
        return hash(tuple(sorted(hash((k, v)) for k, v in self._items())))

    def __eq__(self, other):
        if not isinstance(other, SlurmOptions):
            return False
        return hash(self) == hash(other)

    def as_dict(self) -> typing.Dict:
        return dict(self._items())

    def add_dependencies(self, job_ids, method: str = "afterany"):
        opt = f"{method}:" + ":".join(str(jid) for jid in job_ids)
        if "dependency" in self:
            # There are already dependencies. Trying to extend them.
            if isinstance(self["dependency"], dict):
                dependecy_dict = self["dependency"]
                if method in dependecy_dict:
                    dependecy_dict[method] += ":" + ":".join(
                        str(jid) for jid in job_ids
                    )
                else:
                    dependecy_dict[method] = ":".join(str(jid) for jid in job_ids)
            elif isinstance(self["dependency"], str):
                self["dependency"] += "," + opt
            else:
                # Could not extend dependencies because I have no idea what is going on.
                msg = "Key 'dependency' has unexpected type."
                raise RuntimeError(msg)
        else:
            self["dependency"] = opt



import pytest

import slurminade
from slurminade.guard import TooManyDispatchesError, _DispatchGuard, set_dispatch_limit


@slurminade.slurmify()
def f():
    pass


def test_dispatch_guard_simple():
    slurminade.set_entry_point(__file__)
    dg = _DispatchGuard(3)
    dg()
    dg()
    dg()
    with pytest.raises(TooManyDispatchesError):
        dg()


def test_dispatch_guard_dispatch_limit():
    slurminade.set_entry_point(__file__)
    set_dispatch_limit(3)
    f.distribute()
    f.distribute()
    f.distribute()
    with pytest.raises(TooManyDispatchesError):
        f.distribute()


def test_dispatch_guard_dispatch_limit_batch():
    slurminade.set_entry_point(__file__)
    set_dispatch_limit(2)
    with slurminade.JobBundling(max_size=2):
        for _ in range(4):
            f.distribute()
    with pytest.raises(TooManyDispatchesError):
        f.distribute()



"""
Contains code for bundling function calls together.
"""

import logging
import typing
from collections import defaultdict
from pathlib import Path

from .dispatcher import (
    Dispatcher,
    FunctionCall,
    get_dispatcher,
    set_dispatcher,
)
from .function import SlurmFunction
from .job_reference import JobReference
from .options import SlurmOptions


class BundlingJobReference(JobReference):
    def __init__(self) -> None:
        super().__init__()

    def get_job_id(self) -> typing.Optional[int]:
        return None

    def get_exit_code(self) -> typing.Optional[int]:
        return None

    def get_info(self) -> typing.Dict[str, typing.Any]:
        return {}


class TaskBuffer:
    """
    A simple container to buffer all the tasks by their options.
    We can only bundle tasks with the same slurm options.
    """

    def __init__(self):
        self._tasks = defaultdict(list)

    def add(self, task: FunctionCall, options: SlurmOptions, entry_point: Path) -> int:
        self._tasks[(entry_point, options)].append(task)
        return len(self._tasks[(entry_point, options)])

    def items(self):
        for (entry_point, opt), tasks in self._tasks.items():
            if tasks:
                yield entry_point, opt, tasks

    def clear(self):
        self._tasks.clear()


class JobBundling(Dispatcher):
    """
    The logic to buffer the function calls. It wraps the original dispatcher.

    You can use::

        with slurminade.Batch(max_size=20) as batch:  # automatically bundles up to 20 tasks
            # run 100x f
            for i in range(100):
                f.distribute(i)

    to automatically bundle up to 20 tasks and distribute them.
    """

    def __init__(self, max_size: int):
        """
        :param max_size: Bundle up to this many calls.
        """
        super().__init__()
        self.max_size = max_size
        self.subdispatcher = get_dispatcher()
        self._tasks = TaskBuffer()
        self._all_job_refs = []

    def flush(self) -> typing.List[JobReference]:
        """
        Distribute all buffered tasks. Return the jobs used.
        This method is called automatically when the context is exited.
        However, you may want to call it manually to get the job references,
        for example to use them for dependency management with ``wait_for``.
        :param options: Only flush tasks with specific options.
        :return: A list of job references.
        """
        job_refs = []
        for entry_point, opt, tasks_ in self._tasks.items():
            tasks = tasks_
            while tasks:
                job_ref = self.subdispatcher(tasks[: self.max_size], opt, entry_point)
                job_refs.append(job_ref)
                tasks = tasks[self.max_size :]
        self._tasks.clear()
        self._all_job_refs.extend(job_refs)
        return job_refs

    def get_all_job_ids(self) -> typing.List[int]:
        """
        Return all job ids that have been used.
        """
        job_ids = [job_ref.get_job_id() for job_ref in self._all_job_refs]
        return [jid for jid in job_ids if jid is not None]

    def get_all_jobs(self) -> typing.List[JobReference]:
        return list(self._all_job_refs)

    def add(self, func: SlurmFunction, *args, **kwargs):
        """
        You can also add a task using `add` instead of `distribute`.
        :param func: The function to call
        :param args: The positional arguments
        :param kwargs: The keywords arguments.
        :return: None
        """
        self._dispatch(
            [FunctionCall(func.func_id, args, kwargs)],
            func.special_slurm_opts,
            func.get_entry_point(),
        )

    def _dispatch(
        self,
        funcs: typing.Iterable[FunctionCall],
        options: SlurmOptions,
        entry_point: Path,
        block: bool = False,
    ) -> JobReference:
        if block:
            # if blocking, we don't buffer, but dispatch immediately
            return self.subdispatcher._dispatch(funcs, options, entry_point, block=True)
        for func in funcs:
            self._tasks.add(func, options, entry_point)
        return BundlingJobReference()

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> JobReference:
        conf = SlurmOptions(conf if conf else {})
        return self.subdispatcher.srun(command, conf, simple_slurm_kwargs)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> JobReference:
        conf = SlurmOptions(conf if conf else {})
        return self.subdispatcher.sbatch(command, conf, simple_slurm_kwargs)

    def __enter__(self):
        self.subdispatcher = get_dispatcher()
        set_dispatcher(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logging.getLogger("slurminade").error("Aborted due to exception.")
            return
        self.flush()
        set_dispatcher(self.subdispatcher)

    def _log_dispatch(self, funcs: typing.List[FunctionCall], options: SlurmOptions):
        if len(funcs) == 1:
            logging.getLogger("slurminade").info(
                f"Adding task to batch with options {options}: {funcs[0]}"
            )
        else:
            logging.getLogger("slurminade").info(
                f"Adding {len(funcs)} tasks to batch with options {options}: {', '.join([str(f) for f in funcs])}"
            )

    def __del__(self):
        self.flush()

    def join(self):
        self.flush()
        return self.subdispatcher.join()

    def is_sequential(self):
        return self.subdispatcher.is_sequential()


class Batch(JobBundling):
    """
    Compatibility alias for JobBundling. This is the old name. Deprecated.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logging.getLogger("slurminade").warning(
            "The `Batch` class has been renamed to `JobBundling`. Please update your code."
        )



import abc
from typing import Any, Dict, Optional


class JobReference(abc.ABC):
    """
    The reference to a job that has been dispatched using any dispatcher.
    While the primary purpose of slurminde is to dispatch jobs to slurm,
    return a reference to a slurm job, there are other dispatchers for
    local testing. The dispatchers can create their own job reference
    classes, such that they can add additional information to the job
    as needed.
    """

    @abc.abstractmethod
    def get_job_id(self) -> Optional[int]:
        pass

    @abc.abstractmethod
    def get_exit_code(self) -> Optional[int]:
        pass

    @abc.abstractmethod
    def get_info(self) -> Dict[str, Any]:
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(job_id={self.get_job_id()}, exit_code={self.get_exit_code()}, info={self.get_info()})"



import os
from pathlib import Path

import slurminade

f_file = "./f_test_file.txt"
g_file = "./g_test_file.txt"


@slurminade.slurmify()
def f():
    with Path(f_file).open("w") as file:
        file.write("test")


def delete_f():
    Path(f_file).unlink(missing_ok=True)


@slurminade.slurmify()
def g(x, y):
    with Path(g_file).open("w") as file:
        file.write(f"{x}:{y}")


def delete_g():
    Path(g_file).unlink(missing_ok=True)


def get_file_name():
    filename = os.getenv("PYTEST_CURRENT_TEST")  # doesn't seem to work always
    filename = filename.split("::")[0] if filename else __file__  # workaround
    if not Path(filename).exists():  # sometimes the test folder gets duplicated.
        filename = "/".join(filename.split("/")[1:])
    return Path(filename).resolve()


def test_subprocess_1():
    slurminade.set_dispatcher(slurminade.SubprocessDispatcher())
    slurminade.set_dispatch_limit(100)
    slurminade.set_entry_point(get_file_name())

    delete_f()
    f.distribute()
    assert Path(f_file).exists()
    delete_f()


def test_subprocess_2():
    slurminade.set_dispatcher(slurminade.SubprocessDispatcher())
    slurminade.set_entry_point(get_file_name())
    slurminade.set_dispatch_limit(100)

    delete_g()
    g.distribute(x="a", y=2)
    assert Path(g_file).exists()
    with Path(g_file).open() as file:
        assert file.readline() == "a:2"
    delete_g()



"""
The internal datastructure to save all the slurmified functions.
Not relevant for endusers.
"""

import inspect
import logging
import pathlib
import typing
from pathlib import Path
from typing import Optional

from .execute_cmds import call_slurminade_to_get_function_ids


class FunctionMap:
    """
    The function map assigns functions an id and stores them to be called later.
    This id is reproducible such that the slurm node can retrieve the function
    it is supposed to call just by the id.
    """

    # Needed to save the entry point at the slurm node.
    # The slurm node just executes the file content of a script, so the file name is lost.
    # slurminade will set this value in the beginning to reconstruct it.
    entry_point: typing.Optional[str] = None
    _data: typing.ClassVar[typing.Dict[str, typing.Callable]] = {}
    _ids: typing.ClassVar[Optional[typing.Set[str]]] = set()

    @staticmethod
    def get_id(func: typing.Callable) -> str:
        """
        Returns the unique id of a function. Necessary to slurmify it.
        Probably not needed by the enduser.
        It uses the function name and its file, which should be sufficient unless you
        do not overwrite a function (which is bad anyway).
        :param func: The function you want the id of.
        :return: The id as string.
        """
        file = inspect.getfile(func)
        if file == "<string>":  # on the slurm node, the functions in the entry point
            # are named `<string>`.
            if not FunctionMap.entry_point:
                msg = "No entry point known."
                raise RuntimeError(msg)
            file = FunctionMap.entry_point
        path = Path(file).resolve()
        return f"{path}:{func.__name__}"

    @staticmethod
    def get_readable_name(func_id: str) -> str:
        return func_id.split(":")[-1]

    @staticmethod
    def check_compatibility(func: typing.Callable):
        """
        Throw if the function cannot be assigned an id.
        :param func: The function to be checked.
        :return: None
        """
        if (
            not func.__name__
            or func.__name__ == "<lambda>"
            or not inspect.getfile(func)
        ):
            msg = "Can only slurmify proper functions."
            raise ValueError(msg)

    @staticmethod
    def register(func: typing.Callable, allow_overwrite: bool = False) -> str:
        """
        Register a function, allowing it to be called just by its id.
        :param func: The function to be stored. Needs to be a proper function.
        :return: The function's id.
        """
        FunctionMap.check_compatibility(func)
        func_id = FunctionMap.get_id(func)
        if func_id in FunctionMap._data and not allow_overwrite:
            msg = "Multiple function definitions!"
            raise RuntimeError(msg)
        FunctionMap._data[func_id] = func
        return func_id

    @staticmethod
    def exists(func_id: str) -> bool:
        return func_id in FunctionMap._data

    @staticmethod
    def call(
        func_id: str, args: typing.Iterable, kwargs: typing.Dict[str, typing.Any]
    ) -> typing.Any:
        """
        Calls a function by its id.
        :param func_id: The id of the function to be called.
        :param args: The positional arguments.
        :param kwargs: The keyword arguments.
        :return: The return value of the function.
        """
        if func_id not in FunctionMap._data:
            msg = f"Function '{func_id}' unknown!"
            raise KeyError(msg)
        return FunctionMap._data[func_id](*args, **kwargs)

    @staticmethod
    def check_id(func_id: str, entry_point: Path) -> bool:
        if FunctionMap._ids is None:
            return True
        if func_id in FunctionMap._ids:
            return True
        try:
            # Try to identify the available function ids when the entry point is known.
            # This can help detect some bugs early on.
            FunctionMap._ids = call_slurminade_to_get_function_ids(entry_point)
        except Exception as e:
            logging.getLogger("slurminade").warning(
                "Cannot verify function ids before submitting to slurm: %s. This is not critical, things will just be more difficult to debug in case you make an error.",
                e,
            )
            FunctionMap._ids = None
            return True
        logging.getLogger("slurminade").info(
            "Entry point '%s' has functions %s",
            entry_point,
            list(FunctionMap._ids),
        )
        return func_id in FunctionMap._ids

    @staticmethod
    def get_all_ids() -> typing.List[str]:
        return list(FunctionMap._data.keys())


def set_entry_point(entry_point: typing.Union[str, pathlib.Path]) -> None:
    """
    This function usually is not necessary for endusers.
    Set a manual entry point. This can allow you to use slurmify from the interactive
    interpreter.
    :param entry_point: A path to the entry point file.
    :return: None
    """
    entry_point = Path(entry_point)
    if not entry_point.is_file() or not str(entry_point).endswith(".py"):
        msg = f"Illegal entry point ({entry_point})."
        raise ValueError(msg)
    entry_point = entry_point.resolve()
    FunctionMap.entry_point = str(entry_point)
    # SlurmFunction.dispatcher.entry_point = entry_point


def get_entry_point() -> Path:
    if FunctionMap.entry_point is None:
        import __main__

        # check if attribute __file__ is available
        if not hasattr(__main__, "__file__"):
            msg = "No entry point known."
            raise FileNotFoundError(msg)

        entry_point = __main__.__file__
        if not Path(entry_point).is_file() or Path(entry_point).suffix != ".py":
            msg = "No entry point known."
            raise FileNotFoundError(msg)

        set_entry_point(entry_point)
    assert FunctionMap.entry_point is not None
    path = Path(FunctionMap.entry_point)
    if not path.exists():
        msg = f"Entry point {path} does not exist."
        raise FileNotFoundError(msg)
    return path



from pathlib import Path

import slurminade

test_file_path = Path("./f_test_file.txt")


@slurminade.node_setup
def f():
    with test_file_path.open("w") as file:
        file.write("node_setup")


@slurminade.slurmify
def nil():
    pass


def test_node_setup():
    slurminade.set_entry_point(__file__)
    test_file_path.unlink(missing_ok=True)
    dispatcher = slurminade.SubprocessDispatcher()
    slurminade.set_dispatcher(dispatcher)
    slurminade.set_dispatch_limit(100)
    nil.distribute()
    with test_file_path.open() as file:
        assert file.readline() == "node_setup"
    if test_file_path.exists():  # delete the file
        test_file_path.unlink()



import inspect
import logging
import subprocess
import typing
from enum import Enum
from pathlib import Path

from .dispatcher import FunctionCall, dispatch, get_dispatcher
from .function_map import FunctionMap, get_entry_point
from .guard import guard_recursive_distribution
from .job_reference import JobReference
from .options import SlurmOptions


class CallPolicy(Enum):
    """
    Policy for the call of a function.
    """

    LOCALLY = 0
    DISTRIBUTED = 1
    DISTRIBUTED_BLOCKING = 2


class SlurmFunction:
    """
    A wrapper around a function that allows it to be distributed to slurm.

    .. code-block:: python

        @slurmify(function specific slurm options)
        def f(foobar):
            print(foobar)

        if __name__=="__main__":
            assert isinstance(f, SlurmFunction), "f has become a SlurmFunction"
            jid = f.distribute("hello")
            f.wait_for(jid).distribute("bye")

    """

    def __init__(
        self,
        special_slurm_opts: typing.Dict,
        func: typing.Callable,
        func_id: str,
        call_policy: CallPolicy = CallPolicy.LOCALLY,
    ):
        self.special_slurm_opts = SlurmOptions(**special_slurm_opts)
        self.func = func
        self.func_id = func_id
        self.call_policy = call_policy
        self.defining_file = Path(inspect.getfile(func))

    def update_options(self, conf: typing.Dict[str, typing.Any]):
        self.special_slurm_opts.update(conf)

    def wait_for(
        self,
        job_ids: typing.Union[JobReference, typing.Iterable[JobReference]],
        method: str = "afterany",
    ) -> "SlurmFunction":
        """
        Add a dependency to a distribution.
        `f_jid = f.wait_for(job_ids).distribute("hello")`
        f will only be executed once all jobs within job_ids have finished (or failed).
        Its own job id can then also be used as dependency for other distributions.
        Note that this method does not manipulate the original function but returns
        a new function object, so you have to use chaining.
        Not every dispatcher returns valid job ids on distribute. For example with
        `Batch`, you get the job ids with `job_ids = batch.flush()`.
        :param job_ids: A single job id or an iterable of job ids.
        :param method: 'after'|'afterany'|'afternotok'|'afterok'|'singleton'
        :return: Chainable slurm function object.
        """
        sfunc = SlurmFunction(self.special_slurm_opts, self.func, self.func_id)
        job_ids = (
            [job_ids] if isinstance(job_ids, JobReference) else list(job_ids)
        )  # make sure it is a list
        if not job_ids and not get_dispatcher().is_sequential():
            msg = "Creating a dependency on an empty list of job ids."
            msg += " This is probably an error in your code."
            msg += " Maybe you are using `Batch` but flush outside of the `with` block?"
            raise RuntimeError(msg)
        if (
            any(jid.get_job_id() is None for jid in job_ids)
            and not get_dispatcher().is_sequential()
        ):
            msg = "Invalid job id. Not every dispatcher can directly return job ids, because it may not directly distribute them or doesn't distribute them at all."
            raise RuntimeError(msg)
        sfunc.special_slurm_opts.add_dependencies(
            [jid.get_job_id() for jid in job_ids], method
        )
        return sfunc

    def with_options(self, **kwargs) -> "SlurmFunction":
        """
        Add slurm options to the function.
        :param kwargs: The slurm options.
        :return: The modified function.
        """
        sfunc = SlurmFunction(self.special_slurm_opts, self.func, self.func_id)
        sfunc.update_options(kwargs)
        return sfunc

    def _check(self, args, kwargs):
        """
        Check if the arguments match the function signature.
        """
        inspect.signature(self.func).bind(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Direct call of the original function.
        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        :return: The return value of the function.
        """
        if self.call_policy == CallPolicy.LOCALLY:
            return self.run_locally(*args, **kwargs)
        if self.call_policy == CallPolicy.DISTRIBUTED:
            return self.distribute(*args, **kwargs)
        if self.call_policy == CallPolicy.DISTRIBUTED_BLOCKING:
            return self.distribute_and_wait(*args, **kwargs)
        msg = "Unknown call policy."
        raise RuntimeError(msg)

    def get_entry_point(self) -> Path:
        """
        Returns the entry point for the function.
        Either it is defined in the FunctionMap, or the defining file is used.
        """
        try:
            return get_entry_point()
        except FileNotFoundError:
            logging.getLogger("slurminade").debug(
                "Using defining file %s as entry point.", self.defining_file
            )
            return self.defining_file

    def distribute(self, *args, **kwargs) -> JobReference:
        """
        Try to distribute function call. If slurm is not available, a direct function
        call will be performed.
        `f.distribute("hello")`
        Call with function arguments.
        :return: Job id. Not necessarily valid (usually -1 in this case).
        """
        self._check(args, kwargs)
        guard_recursive_distribution()
        return dispatch(
            [FunctionCall(self.func_id, args, kwargs)],
            self.special_slurm_opts,
            entry_point=self.get_entry_point(),
            block=False,
        )

    def distribute_and_wait(self, *args, **kwargs) -> JobReference:
        """
        Distribute the function and wait for it to finish.
        :param args: The positional arguments.
        :param kwargs: The keyword arguments.
        :return: The job id.
        """
        self._check(args, kwargs)
        guard_recursive_distribution()
        return dispatch(
            [FunctionCall(self.func_id, args, kwargs)],
            self.special_slurm_opts,
            entry_point=self.get_entry_point(),
            block=True,
        )

    def run_locally(self, *args, **kwargs):
        """
        Call the function locally.
        `f.local("hello")`
        Call with function arguments.
        :return: The return value of the function.
        """
        return self.func(*args, **kwargs)

    def __str__(self) -> str:
        return self.func.__name__

    @staticmethod
    def call(func_id, *args, **kwargs):
        return FunctionMap.call(func_id, args, kwargs)


def slurmify(
    f=None, **args
) -> typing.Union[typing.Callable[[typing.Callable], SlurmFunction], SlurmFunction]:
    """
    Decorator: Make a function distributable to slurm.
    Usage:

    .. code-block:: python

        @slurmify()
        def func(a, b):
            pass

    :param f: Function
    :param args: Special slurm options for this function.
    :return: A decorated function, callable with slurm.
    """

    if f:  # use default parameters
        func_id = FunctionMap.register(f)
        return SlurmFunction({}, f, func_id)

    def dec(func) -> SlurmFunction:
        func_id = FunctionMap.register(func)
        return SlurmFunction(args, func, func_id)

    return dec


def _slurmify(
    allow_overwrite: bool, **args
) -> typing.Union[typing.Callable[[typing.Callable], SlurmFunction], SlurmFunction]:
    """
    Decorator: Make a function distributable to slurm.
    Usage:

    .. code-block:: python

        @slurmify_()
        def func(a, b):
            pass

    :param f: Function
    :param args: Special slurm options for this function.
    :return: A decorated function, callable with slurm.
    """

    def dec(func) -> SlurmFunction:
        func_id = FunctionMap.register(func, allow_overwrite=allow_overwrite)
        return SlurmFunction(args, func, func_id)

    return dec


@_slurmify(allow_overwrite=True)
def shell(cmd: typing.Union[str, typing.List[str]]):
    """
    Execute a command.
    :param cmd: The command to be executed.
    """
    subprocess.run(cmd, check=True, shell=True)



"""
The dispatcher distribute function calls to slurm or the local machine.
It can be accessed with `get_dispatcher` and set with `set_dispatcher`.
This allows to change the behaviour of the distribution, e.g., we use it for batch:
Batch simply wraps the dispatcher by a buffered version.
"""

import abc
import logging
import os
import shlex
import shutil
import subprocess
import typing
from pathlib import Path
from typing import Any, Dict, Optional

import simple_slurm

from .conf import _get_conf
from .execute_cmds import create_slurminade_command
from .function_call import FunctionCall
from .function_map import FunctionMap, get_entry_point
from .guard import dispatch_guard
from .job_reference import JobReference
from .options import SlurmOptions

# MAX_ARG_STRLEN on a Linux system with PAGE_SIZE 4096 is 131072
DEFAULT_MAX_ARG_LENGTH = 100000


class Dispatcher(abc.ABC):
    """
    Abstract dispatcher to be inherited by all concrete dispatchers.
    For implementing a dispatcher you have to implement `_dispatch`, `srun` and `sbatch`.

    """

    @abc.abstractmethod
    def _dispatch(
        self,
        funcs: typing.Iterable[FunctionCall],
        options: SlurmOptions,
        entry_point: Path,
        block: bool = False,
    ) -> JobReference:
        """
        Define how to dispatch a number of function calls.
        :param funcs: The function calls to be dispatched.
        :param options: The slurm options to be used.
        :return: The job id. Use -1 if not applicable (e.g., because buffered)
        """

    @abc.abstractmethod
    def srun(
        self,
        command: str,
        conf: typing.Optional[SlurmOptions] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> JobReference:
        """
        Define how you want to execute an `srun` command. This command is directly
        executed and only terminates after completion.
        :param command: A system command, e.g. `echo hello world > foobar.txt`.
        :param conf: The slurm configuration.
        :param simple_slurm_kwargs: Additional options for simple_slurm.
        :return: Job id
        """

    @abc.abstractmethod
    def sbatch(
        self,
        command: str,
        conf: typing.Optional[SlurmOptions] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> JobReference:
        """
        Define how you want to execute an `sbatch` command. The command is scheduled
        and the function return immediately.
        :param command: A system command, e.g. `echo hello world > foobar.txt`.
        :param conf: The slurm configuration.
        :param simple_slurm_kwargs: Additional options for simple_slurm.
        :return: Job id.
        """

    def _log_dispatch(self, funcs: typing.List[FunctionCall], options: SlurmOptions):
        if len(funcs) == 1:
            logging.getLogger("slurminade").info(
                f"Dispatching task with options {options}: {funcs[0]}"
            )
        else:
            logging.getLogger("slurminade").info(
                f"Dispatching task consisting of {len(funcs)} function calls with options {options}: {', '.join([str(f) for f in funcs])}"
            )

    def __call__(
        self,
        funcs: typing.Union[FunctionCall, typing.Iterable[FunctionCall]],
        options: SlurmOptions,
        entry_point: Path,
        block: bool = False,
    ) -> JobReference:
        """
        Dispatches a function call or a number of function calls.
        :param funcs: The function calls to be distributed.
        :param options: The slurm options to be used.
        :return: Job id.
        """
        if isinstance(funcs, FunctionCall):
            funcs = [funcs]
        funcs = list(funcs)
        self._log_dispatch(funcs, options)
        return self._dispatch(funcs, options, entry_point, block)

    def is_sequential(self):
        """
        Return true if the dispatcher works sequential. In this case, the dependencies
        are trivially fulfilled. Slurm does not work sequentially, because this
        would destroy its purpose. In some cases however, you do not want to use
        slurm for compatibility reasons, without changing the script. In these cases,
        this function tells slurminade not to be too strict about dependencies.
        :return: True is tasks are executed sequentially, false if not.
        """
        return False

    def join(self):
        if self.is_sequential():
            # Already sequential, nothing to do
            return
        msg = "Joining is not implemented for this dispatcher."
        raise NotImplementedError(msg)


class TestJobReference(JobReference):
    def get_job_id(self) -> None:
        return None

    def get_exit_code(self) -> None:
        return None

    def get_info(self) -> Dict[str, Any]:
        return {"info": "test"}


class TestDispatcher(Dispatcher):
    """
    A dummy dispatcher that just prints the output. Primarily for debugging and testing.
    """

    def __init__(self):
        super().__init__()
        self.calls = []
        self.sbatches = []
        self.sruns = []
        self.max_arg_length = DEFAULT_MAX_ARG_LENGTH

    def _dispatch(
        self,
        funcs: typing.Iterable[FunctionCall],
        options: SlurmOptions,  # noqa: ARG002
        entry_point: Path,  # noqa: ARG002
        block: bool = False,  # noqa: ARG002
    ) -> JobReference:
        dispatch_guard()
        funcs = list(funcs)
        command = create_slurminade_command(
            get_entry_point(), funcs, self.max_arg_length
        )
        logging.getLogger("slurminade").info(command)
        self.calls.append(funcs)
        self._cleanup(command)
        return TestJobReference()

    def _cleanup(self, command):
        args = shlex.split(command)
        if args[-2] != "temp":
            return
        filename = Path(args[-1])
        if filename.exists():
            filename.unlink()

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,  # noqa: ARG002
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,  # noqa: ARG002
    ):
        dispatch_guard()
        self.sruns.append(command)
        logging.getLogger("slurminade").info("[test output] SRUN %s", command)
        return TestJobReference()

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,  # noqa: ARG002
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,  # noqa: ARG002
    ):
        dispatch_guard()
        self.sbatches.append(command)
        logging.getLogger("slurminade").info("[test output] SBATCH %s", command)
        return TestJobReference()

    def is_sequential(self):
        return True


class SlurmJobReference(JobReference):
    def __init__(self, job_id, exit_code, mode: str):
        self.job_id = job_id
        self.exit_code = exit_code
        self.mode = mode

    def get_job_id(self) -> int:
        return self.job_id

    def get_exit_code(self) -> Optional[int]:
        return self.exit_code

    def get_info(self) -> Dict[str, Any]:
        return {
            "job_id": self.job_id,
            "exit_code": self.exit_code,
            "on_slurm": True,
            "mode": self.mode,
        }


class SlurmDispatcher(Dispatcher):
    """
    The most important dispatcher: Distributing function calls to slurm.
    """

    def __init__(self):
        super().__init__()
        if not shutil.which("sbatch"):
            msg = "Slurm could not be found."
            raise RuntimeError(msg)
        self.max_arg_length = DEFAULT_MAX_ARG_LENGTH
        self._all_job_ids = []
        self._join_dependencies = []

    def _create_slurm_api(self, special_slurm_opts):
        conf = _get_conf(special_slurm_opts)
        return simple_slurm.Slurm(**conf)

    def _job_name(self, funcs: typing.List[FunctionCall]) -> str:
        func_names = list({FunctionMap.get_readable_name(f.func_id) for f in funcs})
        if len(funcs) == 1:
            return f"slurminade:{func_names[0]}"
        return f"slurminade[batch]:{func_names[0]}..."

    def _dispatch(
        self,
        funcs: typing.Iterable[FunctionCall],
        options: SlurmOptions,
        entry_point: Path,
        block: bool = False,
    ) -> SlurmJobReference:
        dispatch_guard()
        if "job_name" not in options:
            funcs = list(funcs)
            # This is complicated to prevent warnings about the type
            options = SlurmOptions(**options.as_dict())
            options["job_name"] = self._job_name(funcs)
        options = SlurmOptions(**options)
        if self._join_dependencies:
            options.add_dependencies(self._join_dependencies, "afterany")
        slurm = self._create_slurm_api(options)
        command = create_slurminade_command(entry_point, funcs, self.max_arg_length)
        logging.getLogger("slurminade").debug(command)
        if block:
            ret = slurm.srun(command)
            logging.getLogger("slurminade").info(
                "Returned from srun with exit code %s", ret
            )
            return SlurmJobReference(None, ret, "srun")
        jid = slurm.sbatch(command)
        self._all_job_ids.append(jid)
        return SlurmJobReference(jid, None, "sbatch")

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> SlurmJobReference:
        dispatch_guard()
        conf = _get_conf(conf)
        slurm = simple_slurm.Slurm(**conf)
        logging.getLogger("slurminade").debug("SBATCH %s", command)
        if simple_slurm_kwargs:
            jid = slurm.sbatch(command, **simple_slurm_kwargs)
        else:
            jid = slurm.sbatch(command)
        self._all_job_ids.append(jid)
        return SlurmJobReference(jid, None, "sbatch")

    def join(self):
        if not self._all_job_ids:
            return
        self._join_dependencies = list(set(self._all_job_ids))

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> SlurmJobReference:
        dispatch_guard()
        conf = _get_conf(conf)
        slurm = simple_slurm.Slurm(**conf)
        logging.getLogger("slurminade").debug("SRUN %s", command)
        if simple_slurm_kwargs:
            ret = slurm.srun(command, **simple_slurm_kwargs)
        else:
            ret = slurm.srun(command)
        return SlurmJobReference(None, ret, "srun")


class SubprocessJobReference(JobReference):
    def __init__(self):
        pass

    def get_job_id(self) -> Optional[int]:
        return None

    def get_exit_code(self) -> Optional[int]:
        return None

    def get_info(self) -> Dict[str, Any]:
        return {"on_slurm": False}


class SubprocessDispatcher(Dispatcher):
    """
    A dispatcher for debugging that distributes function calls using subprocesses.
    Thus, it uses the same serialization mechanisms, but without a slurm dependency.
    Completely useless for productive purposes. Use `DirectCallDispatcher` if you
    don't want to use slurm.
    Despite using subprocesses, it does not parallelize but works sequential.
    """

    def __init__(self):
        super().__init__()
        self.max_arg_length = DEFAULT_MAX_ARG_LENGTH

    def _dispatch(
        self,
        funcs: typing.Iterable[FunctionCall],
        options: SlurmOptions,  # noqa: ARG002
        entry_point: Path,
        block: bool = False,  # noqa: ARG002
    ) -> int:
        dispatch_guard()
        command = create_slurminade_command(entry_point, funcs, self.max_arg_length)
        os.system(command)
        return -1

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,  # noqa: ARG002
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,  # noqa: ARG002
    ):
        dispatch_guard()
        logging.getLogger("slurminade").debug("SRUN %s", command)
        subprocess.run(command, check=True)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,  # noqa: ARG002
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,  # noqa: ARG002
    ):
        self.srun(command)

    def is_sequential(self):
        return True


class LocalJobReference(JobReference):
    def get_job_id(self) -> None:
        return None

    def get_exit_code(self) -> None:
        return None

    def get_info(self) -> Dict[str, Any]:
        return {"on_slurm": False}


class DirectCallDispatcher(Dispatcher):
    """
    A dispatcher that calls functions as if we would not use slurminade.
    This allows compatibility of scripts also on computers not integrated into
    the slurm network.
    """

    def _dispatch(
        self,
        funcs: typing.Iterable[FunctionCall],
        options: SlurmOptions,  # noqa: ARG002
        entry_point: Path,  # noqa: ARG002
        block: bool = False,  # noqa: ARG002
    ) -> LocalJobReference:
        dispatch_guard()
        for func in funcs:
            FunctionMap.call(func.func_id, func.args, func.kwargs)
        return LocalJobReference()

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,  # noqa: ARG002
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,  # noqa: ARG002
    ):
        dispatch_guard()
        subprocess.run(command, check=True)
        return LocalJobReference()

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,  # noqa: ARG002
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,  # noqa: ARG002
    ):
        return self.srun(command)

    def is_sequential(self):
        return True


# The current dispatcher. Use with `get_dispatcher` and `set_dispatcher`.
__dispatcher: typing.Optional[Dispatcher] = None


def get_dispatcher() -> Dispatcher:
    """
    Returns the current dispatcher. Creates a dispatcher if none is available.
    First tries to create the slurm-dispatcher (as this is the primary purpose of
    slurminade). If no slurm-environment can be found, it creates a DirectCallDispatcher
    to allow compatibility.
    :return: The dispatcher.
    """
    global __dispatcher  # noqa: PLW0603
    if __dispatcher is None:
        try:
            __dispatcher = SlurmDispatcher()
        except RuntimeError as re:
            logging.getLogger("slurminade").warning(str(re))
            logging.getLogger("slurminade").warning("Using direct calls.")
            __dispatcher = DirectCallDispatcher()
    return __dispatcher


def set_dispatcher(dispatcher: Dispatcher) -> None:
    """
    Replaces the dispatcher. Can be used to enforce a specific dispatcher.
    :param dispatcher: The dispatcher to be used.
    :return: None
    """
    global __dispatcher  # noqa: PLW0603
    __dispatcher = dispatcher
    assert dispatcher == get_dispatcher()


def dispatch(
    funcs: typing.Union[FunctionCall, typing.Iterable[FunctionCall]],
    options: SlurmOptions,
    entry_point: Path,
    block: bool = False,
) -> JobReference:
    """
    Distribute function calls with the current dispatcher.
    :param funcs: The functions calls to be distributed.
    :param options: The slurm options to be used.
    :return: The job id.
    """
    funcs = list(funcs) if not isinstance(funcs, FunctionCall) else [funcs]
    for func in funcs:
        if not FunctionMap.check_id(func.func_id, entry_point):
            msg = f"Function '{func.func_id}' cannot be called from the given entry point."
            raise KeyError(msg)
    return get_dispatcher()(funcs, options, entry_point, block)


def srun(
    command: typing.Union[str, typing.List[str]],
    conf: typing.Union[SlurmOptions, typing.Dict, None] = None,
    simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
) -> JobReference:
    """
    Call srun with the current dispatcher. This command is directly
    executed and only terminates after completion.
    :param command: A system command, e.g. `echo hello world > foobar.txt`.
    :param conf: The slurm configuration.
    :param simple_slurm_kwargs: Additional options for simple_slurm.
    :return: Job id
    """
    if not isinstance(conf, SlurmOptions):
        if conf is None:
            conf = {}
        conf = SlurmOptions(**conf)
    command = (
        command
        if isinstance(command, str)
        else " ".join(shlex.quote(c) for c in command)
    )
    return get_dispatcher().srun(command, conf, simple_slurm_kwargs)


def sbatch(
    command: typing.Union[str, typing.List[str]],
    conf: typing.Union[SlurmOptions, typing.Dict, None] = None,
    simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
) -> JobReference:
    """
    The command is scheduled and the function returns immediately.
    :param command: A system command, e.g. `echo hello world > foobar.txt`.
    :param conf: The slurm configuration.
    :param simple_slurm_kwargs: Additional options for simple_slurm.
    :return: Job id.
    """
    if not isinstance(conf, SlurmOptions):
        if conf is None:
            conf = {}
        conf = SlurmOptions(**conf)
    command = (
        command
        if isinstance(command, str)
        else " ".join(shlex.quote(c) for c in command)
    )
    return get_dispatcher().sbatch(command, conf, simple_slurm_kwargs)


def join():
    """
    Join all jobs that have been dispatched so far.
    :return: None
    """
    get_dispatcher().join()



from pathlib import Path

import slurminade
from slurminade.function import SlurmFunction

f_file = "./f_test_file.txt"
g_file = "./g_test_file.txt"


@slurminade.slurmify()
def f():
    with Path(f_file).open("w") as file:
        file.write("test")


def delete_f():
    Path(f_file).unlink(missing_ok=True)


@slurminade.slurmify()
def g(x, y):
    with Path(g_file).open("w") as file:
        file.write(f"{x}:{y}")


def delete_g():
    Path(g_file).unlink(missing_ok=True)


def test_local_1():
    delete_f()
    SlurmFunction.call(f.func_id)
    assert Path(f_file).exists()
    delete_f()


def test_local_2():
    delete_g()
    SlurmFunction.call(g.func_id, x="a", y=2)

    assert Path(g_file).exists()
    with Path(g_file).open() as file:
        assert file.readline() == "a:2"
    delete_g()



"""
This module provides the starting point for the slurm node. You do not have to call
anything of this file yourself.
"""

import json
import logging
from pathlib import Path

import click

from .function import SlurmFunction
from .function_map import FunctionMap, set_entry_point
from .guard import prevent_distribution
from .node_setup import disable_setup


@click.command()
@click.option(
    "--root",
    type=click.Path(exists=True),
    help="The root file of the task.",
    required=True,
)
@click.option("--calls", type=str, help="The function calls.", required=False)
@click.option(
    "--fromfile",
    type=click.Path(exists=True),
    help="The file to read the function calls from.",
    required=False,
)
@click.option(
    "--listfuncs",
    help="List all available functions.",
    default=False,
    is_flag=True,
    required=False,
)
def main(root, calls, fromfile, listfuncs):
    prevent_distribution()  # make sure, the code on the node does not distribute itself.
    if listfuncs:
        disable_setup()
    set_entry_point(root)
    with Path(root).open() as f:
        code = "".join(f.readlines())

    glob = dict(globals())
    glob["__file__"] = root
    glob["__name__"] = None
    exec(code, glob)

    if listfuncs:
        print(json.dumps(FunctionMap.get_all_ids()))  # noqa: T201
        return
    if calls:
        function_calls = json.loads(calls)
    elif fromfile:
        with Path(fromfile).open() as f:
            logging.getLogger("slurminade").info(
                f"Reading function calls from {fromfile}."
            )
            function_calls = json.load(f)
        Path(fromfile).unlink()
    else:
        msg = "No function calls provided."
        raise ValueError(msg)
    if not isinstance(function_calls, list):
        msg = "Expected a list of function calls."
        raise ValueError(msg)
    # Execute the functions
    for fc in function_calls:
        SlurmFunction.call(fc["func_id"], *fc.get("args", []), **fc.get("kwargs", {}))


if __name__ == "__main__":
    main()



"""
This file saves the default configuration for slurm.
"""

import json
import logging
import os.path
import typing
from pathlib import Path

CONFIG_NAME = ".slurminade_default.json"

__default_conf: typing.Dict = {}


def _load_conf(path: Path):
    try:
        if path.is_file():
            with path.open() as f:
                return json.load(f)
        else:
            return {}
    except Exception as e:
        logging.getLogger("slurminade").error(
            f"slurminade could not open default configuration {path}!\n{e!s}"
        )
    return {}


def update_default_configuration(conf=None, **kwargs):
    """
    Adds or updates the default configuration.
    :param conf: A dictionary with the configuration.
    :param kwargs: Configuration parameters. (alternative to giving a dictionary)
    """
    if conf:
        __default_conf.update(conf)
    if kwargs:
        __default_conf.update(kwargs)


def _load_default_conf():
    path = Path.home() / CONFIG_NAME
    update_default_configuration(_load_conf(path))
    if "XDG_CONFIG_HOME" in os.environ:
        path = Path(os.environ["XDG_CONFIG_HOME"]) / "slurminade" / CONFIG_NAME
        update_default_configuration(_load_conf(path))
    update_default_configuration(_load_conf(Path(CONFIG_NAME)))


_load_default_conf()


def set_default_configuration(conf=None, **kwargs):
    """
    Replaces the default configuration.
    This will overwrite the default configuration with the given one.
    :param conf: A dictionary with the configuration.
    :param kwargs: Configuration parameters. (alternative to giving a dictionary)
    """
    __default_conf = {}
    update_default_configuration(conf, **kwargs)


def _get_conf(conf=None):
    conf = conf if conf else {}
    conf_ = __default_conf.copy()
    conf_.update(conf)
    return conf_



import typing


class FunctionCall:
    """
    A function call to be dispatched.
    """

    def __init__(self, func_id: str, args: typing.Tuple, kwargs: typing.Dict):
        self.func_id = func_id  # the function id, as in FunctionMap
        self.args = args  # the positional arguments for the call
        self.kwargs = kwargs  # the keyword arguments for the call

    def to_json(self) -> typing.Dict:
        """
        Convert call to a json object that can be passed to slurm.
        :return: json object.
        """
        return {"func_id": self.func_id, "args": self.args, "kwargs": self.kwargs}

    def __str__(self) -> str:
        """
        Return a printable string representation of the call, useful for logging.
        """

        def arg_to_str(arg):
            if isinstance(arg, str):
                return f"'{arg}'"
            return str(arg)

        short_args = ", ".join(arg_to_str(a) for a in self.args)
        if len(short_args) > 200:
            short_args = short_args[:200] + "..."
        short_kwargs = ", ".join(f"{k}={arg_to_str(v)}" for k, v in self.kwargs.items())
        if len(short_kwargs) > 200:
            short_kwargs = short_kwargs[:200] + "..."
        args = ", ".join(a for a in [short_args, short_kwargs] if a)
        return f"{self.func_id.split(':')[-1]}({args})"



"""
The commands that can be understood by execute.py
"""

import json
import logging
import os
import shlex
import subprocess
import sys
import typing
from pathlib import Path
from tempfile import mkstemp

from .function_call import FunctionCall


def create_slurminade_command(
    entry_point: Path, funcs: typing.Iterable[FunctionCall], max_arg_length: int
) -> str:
    """
    Creates a terminal command that calls the Python module `slurminade.execute` with the
    provided function calls as an argument. If the total length of the function calls
    exceeds the maximum allowed length of a command line argument, a temporary file is
    created to pass the function calls instead.
    :param funcs: The function calls to be dispatched.
    :param max_arg_length: The maximum allowed length of a command line argument.
    :returns: A string representing the command to be executed in the terminal.
    """
    if not entry_point.exists():
        msg = f"Entry point {entry_point} does not exist."
        raise FileNotFoundError(msg)

    command = (
        f"{sys.executable} -m slurminade.execute --root {shlex.quote(str(entry_point))}"
    )

    # Serialize function calls as JSON
    json_calls = json.dumps([f.to_json() for f in funcs])
    serialized_calls = shlex.quote(json_calls)

    if len(serialized_calls) > max_arg_length:
        # The argument is too long, create temporary file for the JSON
        fd, filename = mkstemp(prefix="slurminade_", suffix=".json", text=True, dir=".")
        logging.getLogger("slurminade").info(
            f"Long function calls. Serializing function calls to temporary file {filename}"
        )
        with os.fdopen(fd, "w") as f:
            f.write(json_calls)
        command += f" --fromfile {filename}"
    else:
        command += f" --calls {serialized_calls}"
    return command


def call_slurminade_to_get_function_ids(entry_point: Path) -> typing.Set[str]:
    cmd = [
        sys.executable,
        "-m",
        "slurminade.execute",
        "--root",
        str(entry_point),
        "--listfuncs",
    ]
    out = subprocess.check_output(cmd).decode()
    out = out.strip().split("\n")[-1].strip()
    ids = json.loads(out)
    return set(ids)
